@@ -15,7 +15,7 @@ use App::HTTP_Proxy_IMP::Debug qw(debug $DEBUG $DEBUG_RX);
 use Net::Inspect::Debug qw(%TRACE);
 use Carp 'croak';
 
-our $VERSION = '0.94';
+our $VERSION = '0.941';
 
 # try IPv6 using IO::Socket::IP or IO::Socket::INET6
 # fallback to IPv4 only
@@ -75,6 +75,8 @@ sub start {
 	    }
 	    croak("IMP module $f could not be loaded: $@") if ! $found;
 	    my %args = $args ? $found->str2cfg($args) :();
+	    my @err = $found->validate_cfg(%args);
+	    die "bad config for $found: @err" if @err;
 	    push @mod, $found->new_factory(%args)
 	}
 	$imp_factory = App::HTTP_Proxy_IMP::IMP->new_factory(@mod);
